/*
 * arch/mips/usip/lib/gpio.c
 *
 * (C) 2004 Robert Schwebel, Pengutronix
 *     2006 Ivan Korzakow
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/list.h>
#include <linux/spinlock.h>
#include <linux/device.h>
#include <linux/sysdev.h>
#include <linux/timer.h>
#include <linux/proc_fs.h>
#include <linux/interrupt.h>
#include <linux/gpio.h>

#define DRIVER_NAME "gpio"

static char mapping[255];

MODULE_AUTHOR("John Lenz, Robert Schwebel, Ivan Korzakow");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Generic GPIO Infrastructure");
module_param_string(mapping, mapping, sizeof(mapping), 0);
MODULE_PARM_DESC(mapping,
		 "period delimited options string to map GPIO to userland:\n"
		 "\n"
		 "	<pin>:[out|in][hi|lo]\n"
		 "\n"
		 "	example: mapping=5:out:hi.8:in\n"
	);

static ssize_t gpio_show_level(struct class_device *dev, char *buf);
static ssize_t gpio_store_level(struct class_device *dev,
				const char *buf, size_t size);
static ssize_t gpio_show_policy(struct class_device *dev, char *buf);
irqreturn_t gpio_interrupt(int irq, void *dev_id, struct pt_regs *regs);


struct irq_handler {
	int trigger_type;
	void (*handler)(int gpio_nr, void *private_data);
	void *private_data;
};

struct gpio_properties {
	unsigned int	   pin_nr;
	unsigned char	   policy;	/* GPIO_xxx */
	char		   pin_level;	/* -1=tristate, 0, 1 */
	char		   owner[20];
	struct irq_handler irq;
	struct gpio_device *gpio_dev;
};

struct gpio_device {
	spinlock_t lock;		/* protects the props field */
	struct gpio_properties props;
	struct class_device class_dev;
	struct list_head list;
};

static int irq_counter = 0;
static struct gpio_params *gpio_hw_p;

#define to_gpio_device(d) container_of(d, struct gpio_device, class_dev)

static LIST_HEAD(gpio_list);
static rwlock_t gpio_list_lock = RW_LOCK_UNLOCKED;

/* gpio_device is static, so we don't have to free it here */
static void gpio_class_release(struct class_device *dev)
{
	return;
}

static struct class gpio_class = {
	.name		= "gpio",
	.release	= gpio_class_release,
};


/*
 * Attribute: /sys/class/gpio/gpioX/level
 */
static struct class_device_attribute attr_gpio_level = {
	.attr	= { .name = "level", .mode = 0644, .owner = THIS_MODULE },
	.show	= gpio_show_level,
	.store	= gpio_store_level,
};

static ssize_t gpio_show_level(struct class_device *dev, char *buf)
{
	struct gpio_device *gpio_dev = to_gpio_device(dev);
	ssize_t ret_size = 0;

	if (gpio_dev->props.policy & GPIO_INPUT)
		gpio_dev->props.pin_level = \
			gpio_hw_p->gpio_op->gpio_get_pin_op(gpio_dev->props.pin_nr);

	spin_lock(&gpio_dev->lock);
	ret_size += sprintf(buf, "%i\n", gpio_dev->props.pin_level);
	spin_unlock(&gpio_dev->lock);

	return ret_size;
}

static ssize_t
gpio_store_level(struct class_device *dev, const char *buf, size_t size)
{
	struct gpio_device *gpio_dev = to_gpio_device(dev);
	long value;

	if (gpio_dev->props.policy & GPIO_INPUT)
		return -EINVAL;

	value = simple_strtol(buf, NULL, 10);
	if ((value != 0) && (value != 1))
		return -EINVAL;
	gpio_dev->props.pin_level = value;

	/* set real hardware */
	switch (value) {
	case 0:
		gpio_hw_p->gpio_op->gpio_clear_pin_op(gpio_dev->props.pin_nr);
		gpio_hw_p->gpio_op->gpio_dir_output_op(gpio_dev->props.pin_nr);
		break;
	case 1:
		gpio_hw_p->gpio_op->gpio_set_pin_op(gpio_dev->props.pin_nr);
		gpio_hw_p->gpio_op->gpio_dir_output_op(gpio_dev->props.pin_nr);
		break;
	default: break;
	}
	return size;
}

/*
 * Attribute: /sys/class/gpio/gpioX/policy
 */
static struct class_device_attribute attr_gpio_policy = {
	.attr	= { .name = "policy", .mode = 0444, .owner = THIS_MODULE },
	.show	= gpio_show_policy,
	.store	= NULL,
};

static ssize_t gpio_show_policy(struct class_device *dev, char *buf)
{
	struct gpio_device *gpio_dev = to_gpio_device(dev);
	ssize_t ret_size = 0;

	spin_lock(&gpio_dev->lock);
	if (gpio_dev->props.policy & GPIO_USER)
		ret_size += sprintf(buf,"userspace\n");
	if (gpio_dev->props.policy & GPIO_KERNEL)
		ret_size += sprintf(buf,"kernel\n");
	spin_unlock(&gpio_dev->lock);

	return ret_size;
}

static int gpio_read_proc(char *page, char **start, off_t off,
			  int count, int *eof, void *data)
{
	char *p = page;
	int size = 0;
	struct gpio_device *gpio_dev;
	struct list_head *lact, *ltmp;

	if (off != 0)
		goto end;

	p += sprintf(p, "GPIO	POLICY	     DIRECTION	  OWNER\n");
	list_for_each_safe(lact, ltmp, &gpio_list) {
		gpio_dev = list_entry(lact, struct gpio_device, list);
		p += sprintf(p, "%3i:	", gpio_dev->props.pin_nr);
		if (gpio_dev->props.policy & GPIO_KERNEL)
			p += sprintf(p, "kernel	      ");
		if (gpio_dev->props.policy & GPIO_USER)
			p += sprintf(p, "user space   ");
		if (gpio_dev->props.policy & GPIO_INPUT)
			p += sprintf(p, "input	      ");
		if (gpio_dev->props.policy & GPIO_OUTPUT)
			p += sprintf(p, "output	      ");
		p += sprintf(p, "%s\n", gpio_dev->props.owner);
	}
end:
	size = (p - page);
	if (size <= off + count)
		*eof = 1;
	*start = page + off;
	size -= off;
	if (size > count)
		size = count;
	if (size < 0)
		size = 0;

	return size;
}

/**
 * request_gpio - register a new object of gpio_device class.
 *
 * @pin_nr:	 GPIO pin which is registered
 * @owner:	 name of the driver that owns this pin
 * @policy:	 set policy for this pin, which is one of these:
 *		- GPIO_USER or GPIO_KERNEL
 *		- GPIO_INPUT or GPIO_OUTPUT
 *		For user space registered pins a sysfs entry is added.
 * @init_level: initially configured pin level
 */
int request_gpio(unsigned int pin_nr, const char *owner,
		 unsigned char policy, unsigned char init_level)
{
	int rc, result;
	struct gpio_device *gpio_dev;
	struct list_head *lact, *ltmp;

	if ((pin_nr < gpio_hw_p->gpio_pin_min) ||
	    (pin_nr > gpio_hw_p->gpio_pin_max))
		return -EINVAL;

	list_for_each_safe(lact, ltmp, &gpio_list) {
		gpio_dev = list_entry(lact, struct gpio_device, list);
		if (pin_nr == gpio_dev->props.pin_nr) {
			printk(KERN_ERR "gpio pin %i is already used by %s\n",
			       pin_nr, gpio_dev->props.owner);
			return -EBUSY;
		}
	}

	/* value checks - FIXME: is there a logical xor in C? */
	if ( ( (policy & GPIO_USER) &&	(policy & GPIO_KERNEL)) ||
	     (!(policy & GPIO_USER) && !(policy & GPIO_KERNEL))) {
		printk(KERN_ERR
		       "%s: policy has to be one of GPIO_KERNEL, GPIO_USER\n",
		       DRIVER_NAME);
		return -EINVAL;
	}
	if ( ( (policy & GPIO_INPUT) &&	 (policy & GPIO_OUTPUT)) ||
	     (!(policy & GPIO_INPUT) && !(policy & GPIO_OUTPUT))) {
		printk(KERN_ERR
		       "%s: policy has to be one of GPIO_INPUT, GPIO_OUTPUT\n",
		       DRIVER_NAME);
		return -EINVAL;
	}

	if (list_empty(&gpio_list)) {
		result = gpio_hw_p->gpio_op->gpio_init_specific_op();
		if (result)
			return result;
	}

	gpio_dev = kmalloc(sizeof(struct gpio_device), GFP_KERNEL);

	if (unlikely(!gpio_dev)) {
		printk(KERN_ERR "%s: couldn't allocate memory\n", DRIVER_NAME);
		return -ENOMEM;
	}

	spin_lock_init(&gpio_dev->lock);
	gpio_dev->props.pin_nr = pin_nr;
	gpio_dev->props.policy = policy;
	gpio_dev->props.pin_level = init_level;
	gpio_dev->props.gpio_dev = gpio_dev;
	gpio_dev->props.irq.handler = NULL;
	strncpy(gpio_dev->props.owner, owner, 20);

	memset(&gpio_dev->class_dev, 0, sizeof(gpio_dev->class_dev));
	gpio_dev->class_dev.class = &gpio_class;
	snprintf(gpio_dev->class_dev.class_id, BUS_ID_SIZE, "gpio%i", pin_nr);

	rc = class_device_register(&gpio_dev->class_dev);
	if (unlikely(rc)) {
		printk(KERN_ERR "%s: class registering failed\n", DRIVER_NAME);
		kfree(gpio_dev);
		return rc;
	}

	INIT_LIST_HEAD(&gpio_dev->list);

	/* register the attributes */
	if (policy & GPIO_USER)
		class_device_create_file(&gpio_dev->class_dev,
					 &attr_gpio_level);

	class_device_create_file(&gpio_dev->class_dev, &attr_gpio_policy);

	/* set real hardware */
	if (policy & GPIO_OUTPUT) {
		gpio_hw_p->gpio_op->gpio_dir_output_op(pin_nr);
		switch (init_level) {
		case 0: gpio_hw_p->gpio_op->gpio_clear_pin_op(pin_nr); break;
		case 1: gpio_hw_p->gpio_op->gpio_set_pin_op(pin_nr); break;
		default: break;
		}
	} else if (policy & GPIO_INPUT) {
		gpio_hw_p->gpio_op->gpio_dir_input_op(pin_nr);
	}

	write_lock(&gpio_list_lock);
	list_add_tail(&gpio_dev->list, &gpio_list);
	write_unlock(&gpio_list_lock);

#ifdef GPIO_DEBUG
	printk(KERN_INFO "registered gpio%i\n", pin_nr);
#endif
	return 0;
}
EXPORT_SYMBOL(request_gpio);


/**
 * free_gpio - unregisters a object of gpio_properties class.
 *
 * @pin_nr: pin number to free.
 *
 * Unregisters a previously registered via request_gpio object.
 */
int free_gpio(unsigned int pin_nr)
{
	struct gpio_device *gpio_dev;
	struct list_head *lact, *ltmp;

	if ((pin_nr < gpio_hw_p->gpio_pin_min) ||
	    (pin_nr > gpio_hw_p->gpio_pin_max))
		return -EINVAL;

	list_for_each_safe(lact, ltmp, &gpio_list) {
		gpio_dev = list_entry(lact, struct gpio_device, list);
		if (pin_nr == gpio_dev->props.pin_nr) {

#ifdef GPIO_DEBUG
			printk(KERN_INFO
			       "unregistering gpio pin %i\n", pin_nr);
#endif
			/* Unregister IRQ */
			if (gpio_dev->props.irq.handler != NULL)
				irq_counter--;

			/* unregister attributes */
			if (gpio_dev->props.policy & GPIO_USER) {
				class_device_remove_file(&gpio_dev->class_dev,
							 &attr_gpio_level);
			}

			class_device_remove_file(&gpio_dev->class_dev,
						 &attr_gpio_level);
			class_device_unregister(&gpio_dev->class_dev);
			write_lock(&gpio_list_lock);
			list_del(&gpio_dev->list);
			write_unlock(&gpio_list_lock);
			kfree(gpio_dev);

			/* unregister interrupt source if needed */
			if (irq_counter == 0)
				gpio_hw_p->gpio_op->gpio_free_resource_op();
			return 0;
		}
	}
	/* Not found : nothing to remove ! */
	return 0;
}
EXPORT_SYMBOL(free_gpio);


/**
 * free_all_gpio - unregisters all objects of gpio_properties class.
 */
void free_all_gpios(void)
{
	struct gpio_device *gpio_dev;
	struct list_head *lact, *ltmp;

	list_for_each_safe(lact, ltmp, &gpio_list) {
		gpio_dev = list_entry(lact, struct gpio_device, list);

#ifdef GPIO_DEBUG
		printk(KERN_INFO "unregistering gpio pin %i\n",
		       gpio_dev->props.pin_nr);
#endif
		/* unregister attributes */
		if (gpio_dev->props.policy & GPIO_USER)
			class_device_remove_file(&gpio_dev->class_dev,
						 &attr_gpio_level);

		class_device_remove_file(&gpio_dev->class_dev,
					 &attr_gpio_level);

		class_device_unregister(&gpio_dev->class_dev);
		write_lock(&gpio_list_lock);
		list_del(&gpio_dev->list);
		write_unlock(&gpio_list_lock);
		kfree(gpio_dev);
	}

	if (irq_counter != 0) {
		gpio_hw_p->gpio_op->gpio_free_resource_op();
		irq_counter = 0;
	}

	return;
}
EXPORT_SYMBOL(free_all_gpios);

/**
 * init_gpio_bock - register a block of object of gpio_device class.
 *
 * @first_pin:	Starting pin of the requisted block.
 * @last_pin:	Last pin number requested.
 * @owner:	 name of the driver that owns the block of pin
 * @policy:	 set policy for the block of pin, which is one of these:
 *		- GPIO_USER or GPIO_KERNEL
 *		For user space registered pins a sysfs entry is added.
 * @direction:	Set direction for the block of pin.
 *		 The value is a mapping of the pins :
 *		bit 0 for pin 0, bit 1 for pin 1, etc ...
 * @init_level: initially configured block of pin level (value 0 or 1)
 */
int request_gpio_block(int first_pin, int last_pin, const char *owner,
		       unsigned char policy, unsigned int direction,
		       unsigned char init_level)
{
	int i, result =	 0;

	if ((first_pin <  gpio_hw_p->gpio_pin_min) ||
	    (last_pin > gpio_hw_p->gpio_pin_max) ||
	    first_pin > last_pin)
		return -EINVAL;

	direction = direction >> first_pin;

	/* Register all the pins of the requested block */
	for (i = first_pin ; i <= last_pin ; i++) {
		result = request_gpio(i, owner,
				      policy | (1 << ((direction & 0x1) + 2)),
				      init_level);
		if (result)
			break;
		direction = direction >> 1;
	}

	if (result)
		free_gpio_block(first_pin, i - 1);

	return result;
}
EXPORT_SYMBOL(request_gpio_block);


/**
 * free_gpio_bock - free a block of object of gpio_device class.
 *
 * @first_pin:	Starting pin of the requisted block.
 * @last_pin:	Last pin number requested.
 */
int free_gpio_block(int first_pin, int last_pin)
{
	int i;

	for (i = first_pin; i <= last_pin; i++) {
		free_gpio(i);
	}
	return 0;
}
EXPORT_SYMBOL(free_gpio_block);




/************************************************************************
 *
 * Wrapper for some hardware specific functions
 *
 ************************************************************************/

int gpio_init(void)
{
	return gpio_hw_p->gpio_op->gpio_init_specific_op();
}

int gpio_set_pin(int gpio_nr)
{
	if ((gpio_nr < gpio_hw_p->gpio_pin_min) ||
	    (gpio_nr > gpio_hw_p->gpio_pin_max))
		return -EINVAL;

	return gpio_hw_p->gpio_op->gpio_set_pin_op(gpio_nr);
}

int gpio_clear_pin(int gpio_nr)
{
	if ((gpio_nr < gpio_hw_p->gpio_pin_min) ||
	    (gpio_nr > gpio_hw_p->gpio_pin_max))
		return -EINVAL;

	return gpio_hw_p->gpio_op->gpio_clear_pin_op(gpio_nr);
}

int gpio_dir_input(int gpio_nr)
{
	if ((gpio_nr < gpio_hw_p->gpio_pin_min) ||
	    (gpio_nr > gpio_hw_p->gpio_pin_max))
		return -EINVAL;

	return gpio_hw_p->gpio_op->gpio_dir_input_op(gpio_nr);
}

int gpio_dir_output(int gpio_nr)
{
	if ((gpio_nr < gpio_hw_p->gpio_pin_min) ||
	    (gpio_nr > gpio_hw_p->gpio_pin_max))
		return -EINVAL;

	return gpio_hw_p->gpio_op->gpio_dir_output_op(gpio_nr);
}

int gpio_get_pin(int gpio_nr)
{
	if ((gpio_nr < gpio_hw_p->gpio_pin_min) ||
	    (gpio_nr > gpio_hw_p->gpio_pin_max))
		return -EINVAL;

	return gpio_hw_p->gpio_op->gpio_get_pin_op(gpio_nr);
}

int gpio_is_dir_input(int gpio_nr)
{
	if ((gpio_nr < gpio_hw_p->gpio_pin_min) ||
	    (gpio_nr > gpio_hw_p->gpio_pin_max))
		return -EINVAL;

	return gpio_hw_p->gpio_op->gpio_is_dir_input_op(gpio_nr);
}



/************************************************************************
 *
 * Interrupts
 *
 ************************************************************************/

static void register_linux_interrupt(void)
{
	request_irq(gpio_hw_p->irq_line, gpio_interrupt, SA_INTERRUPT,
		    "gpio_irq", NULL);
}


static void unregister_linux_interrupt(void)
{
	free_irq(gpio_hw_p->irq_line, NULL);
}

/**
 * Register a handler for specific GPIO interrupt.
 *
 * @gpio_nr:	pin number
 * @trigger:	triggering type (level triggerd, edge triggered, ...)
 *		This is hardware dependant, refer to the underlying module.
 * @handler:	the handler to call when interrupts occure
 * @private_data: pointer to some data, or set to NULL
 */
int gpio_request_irq(unsigned int gpio_nr, int trigger,
		     void (*handler)(int gpio_nr, void *private_data),
		     void *private_data)
{
	unsigned int result;
	struct gpio_device *gpio_dev;
	struct list_head *lact, *ltmp;

	if (handler == NULL)
		return -EINVAL;

	if ((gpio_nr < gpio_hw_p->gpio_pin_min) ||
	    (gpio_nr > gpio_hw_p->gpio_pin_max))
		return -EINVAL;

	/* Check that the pin is configured as input */
	if (!gpio_hw_p->gpio_op->gpio_is_dir_input_op(gpio_nr))
		return -EINVAL;

	list_for_each_safe(lact, ltmp, &gpio_list) {
		gpio_dev = list_entry(lact, struct gpio_device, list);
		if (gpio_nr == gpio_dev->props.pin_nr) {
			/* Check that an interrupt is not already registered */
			if (gpio_dev->props.irq.handler != 0)
				return -EBUSY;

			/* Register the handler */
			gpio_dev->props.irq.private_data = private_data;
			gpio_dev->props.irq.trigger_type = trigger;
			gpio_dev->props.irq.handler = handler;

			/* Configure, clear, enable the specified interrupt */
			result = gpio_hw_p->gpio_op->gpio_enable_irq_op(gpio_nr, trigger);
			if (result)
				return result;

			if (irq_counter == 0)
				register_linux_interrupt();

			irq_counter++;
			return 0;
		}
	}
	return 0;
}
EXPORT_SYMBOL(gpio_request_irq);


/**
 * Free an IRQ handler for specific GPIO line.
 *
 * @gpio_nr : pin number
 */
int gpio_free_irq(unsigned int gpio_nr)
{
	struct gpio_device *gpio_dev;
	struct list_head *lact, *ltmp;

	if ((gpio_nr < gpio_hw_p->gpio_pin_min) ||
	    (gpio_nr > gpio_hw_p->gpio_pin_max))
		return -EINVAL;

	list_for_each_safe(lact, ltmp, &gpio_list) {
		gpio_dev = list_entry(lact, struct gpio_device, list);
		if (gpio_nr == gpio_dev->props.pin_nr) {
			/* Check that an interrupt is registered */
			if (gpio_dev->props.irq.handler == NULL)
				return -EINVAL;

			gpio_dev->props.irq.handler = NULL;
			gpio_dev->props.irq.private_data = NULL;
			gpio_hw_p->gpio_op->gpio_disable_irq_op(gpio_nr);

			irq_counter--;
			if (irq_counter == 0)
				unregister_linux_interrupt();

			return 0;
		}
	}
	return 0;
}
EXPORT_SYMBOL(gpio_free_irq);


/**
 * Free all IRQ handler.
 */
int gpio_free_all_irq(void)
{
	int irq_was_used = 0;
	struct gpio_device *gpio_dev;
	struct list_head *lact, *ltmp;

	list_for_each_safe(lact, ltmp, &gpio_list) {
		gpio_dev = list_entry(lact, struct gpio_device, list);
		if (gpio_dev->props.irq.handler != NULL) {
			irq_was_used++;
			gpio_dev->props.irq.handler = NULL;
			gpio_dev->props.irq.private_data = NULL;
		}
	}

	gpio_hw_p->gpio_op->gpio_disable_all_irq_op();
	if (irq_was_used)
		unregister_linux_interrupt();

	irq_counter = 0;

	return 0;
}
EXPORT_SYMBOL(gpio_free_all_irq);



static inline int start_handler(int gpio_nr)
{
	struct gpio_device *gpio_dev;
	struct list_head *lact, *ltmp;

	list_for_each_safe(lact, ltmp, &gpio_list) {
		gpio_dev = list_entry(lact, struct gpio_device, list);
		if (gpio_nr == gpio_dev->props.pin_nr) {
			if (gpio_dev->props.irq.handler == NULL) {
				return -EINVAL;
			}
			gpio_dev->props.irq.handler(gpio_nr,
						    gpio_dev->props.irq.private_data);

			return 0;
		}
	}
	return -EINVAL;
}

/**
 * Call the registered handlers for GPIO interrupts.
 * This function is used as a default gpio dispatcher routine.
 */
irqreturn_t gpio_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	unsigned int gpio_isr;
	int gpio_nr;

	gpio_isr = gpio_hw_p->gpio_op->gpio_get_isr_op();
	while(gpio_isr) {
		if (!gpio_isr)
			break;

		gpio_nr = 0;

		while(gpio_isr) {
			if (gpio_isr & 0x1) {
				/* Acknowledge */
				gpio_hw_p->gpio_op->gpio_ack_irq_op(gpio_nr);
				/* Call handler */
				start_handler(gpio_nr);
			}
			gpio_nr++;
			gpio_isr >>= 1;
		}
		gpio_isr = gpio_hw_p->gpio_op->gpio_get_isr_op();
	}
	return IRQ_HANDLED;
}






/************************************************************************
 *
 * Initialization, setup and exit functions
 *
 ************************************************************************/

static struct sysdev_class gpio_sysclass = {
	set_kset_name("gpio"),
};

static struct sys_device gpio_sys_device = {
	.id	= 0,
	.cls	= &gpio_sysclass,
};

static int gpio_setup(char *s)
{
	char *p, *q = NULL;
	int pin_nr;
	unsigned char policy;
	unsigned char init_level = 0;

	while ((p = strsep(&s, ".,")) != NULL) {

		/* end of command reached? -> next command */
		if (*p == '\0')
			continue;

		/* process one command */
		pin_nr = -1;
		policy = 0;
		q = strsep(&p, ":");
		if (*q == '\0') {
			printk("%s: invalid token (scanning pin_nr): %s\n",
			       DRIVER_NAME, p);
			continue;
		}
		pin_nr = simple_strtoul(q, NULL, 0);
		q = strsep(&p, ":");
		if (*q == '\0') {
			printk("%s: invalid token (scanning direction): %s\n",
			       DRIVER_NAME, p);
			continue;
		}
		policy |= strncmp(q, "in", strlen(q)) == 0 ? GPIO_INPUT : 0;
		policy |= strncmp(q, "out", strlen(q)) == 0 ? GPIO_OUTPUT : 0;
		if (policy == 0) {
			printk("%s: invalid token (scanning policy): %s\n",
			       DRIVER_NAME, p);
			continue;
		}
		if (policy & GPIO_OUTPUT) {
			q = strsep(&p, ":");
			if (*q == '\0') {
				printk("%s: invalid token : %s\n",
				       DRIVER_NAME, p);
				continue;
			}
			init_level = 2;
			if (strncmp(q, "hi", strlen(q)) == 0)
				init_level = 1;
			if (strncmp(q, "lo", strlen(q)) == 0)
				init_level = 0;
			if (init_level == 2) {
				printk("%s: invalid token : %s\n",
				       DRIVER_NAME, p);
				continue;
			}
		}
		policy |= GPIO_USER;
		if (request_gpio(pin_nr, "kernel", policy, init_level) != 0) {
			printk(KERN_ERR
			       "%s: couldn't register GPIO from commandline\n",
			       DRIVER_NAME);
			return -EIO;
		}
	}
	return 0;
}


/*
 * Register the platform dependant functions.
 * This function is called by hardware specific modules
 * to register themselves with this generic layer.
 */
int register_gpio_specific(struct gpio_params *gpio_p)
{
	int result = 0;

	/* Sanity checks */
	if (gpio_p->gpio_op == NULL) {
		result = -EINVAL;
		goto out;
	}

	if (gpio_p->gpio_pin_min > gpio_p->gpio_pin_max) {
		result = -EINVAL;
		goto out;
	}

	/* Register */
	gpio_hw_p = gpio_p;

out:
	return result;
}


static int __init gpio_init_mod(void)
{
	int ret;

	printk(KERN_WARNING "Initialising gpio device class.\n");

	class_register(&gpio_class);

	ret = sysdev_class_register(&gpio_sysclass);
	if (ret) {
		printk(KERN_ERR
		       "%s: couldn't register sysdev class, exiting\n",
		       DRIVER_NAME);
		goto out_sysdev_class;
	}

	ret = sysdev_register(&gpio_sys_device);
	if (ret) {
		printk(KERN_ERR "%s: couldn't register sysdev, exiting\n",
		       DRIVER_NAME);
		goto out_sysdev_register;
	}

	 if (!create_proc_read_entry ("gpio", 0, 0, gpio_read_proc, NULL)) {
		printk(KERN_ERR "%s: couldn't register proc entry, exiting\n",
		       DRIVER_NAME);
		goto out_proc;
	}

	/* parse command line parameters */
	printk(KERN_INFO "%s: mapping=", DRIVER_NAME);
	if (strcmp(mapping,"")) {
		printk("%s\n", mapping);
		if (gpio_setup(mapping) != 0) {
			printk(KERN_ERR
			       "%s: couldn't register 'mapping=..', exiting\n",
			       DRIVER_NAME);
		}
	} else {
		printk("empty.\n");
	}

	return ret;

out_proc:
	sysdev_unregister(&gpio_sys_device);
out_sysdev_register:
	sysdev_class_unregister(&gpio_sysclass);
out_sysdev_class:
	class_unregister(&gpio_class);
	return ret;
}

static void __exit gpio_exit_mod(void)
{
	free_all_gpios();
	return;
}


module_init(gpio_init_mod);
module_exit(gpio_exit_mod);
